from abstra_internals.widgets.widget_base import Input


class PasswordInput(Input):
    type = 'password-input'

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.hint = props.get('hint', None)
        self.required = props.get('required', True)
        self.full_width = props.get('full_width', False)
        self.lowercase_required = props.get('lowercase_required', True)
        self.uppercase_required = props.get('uppercase_required', True)
        self.special_required = props.get('special_required', True)
        self.digit_required = props.get('digit_required', True)
        self.min_length = props.get('min_length', 8)
        self.max_length = props.get('max_length', None)
        self.size = props.get('size', None)
        self.pattern = props.get('pattern', None)
        self.autocomplete = props.get('autocomplete', 'current-password')
        self.placeholder = props.get('placeholder', '')
        self.disabled = props.get('disabled', False)
        self.value = ''
        self.secret = True

    def render(self, ctx: dict):
        return {'value': self.serialize_value(), 'type': self.type, 'key':
            self.key, 'label': self.label, 'hint': self.hint, 'required':
            self.required, 'fullWidth': self.full_width,
            'lowercaseRequired': self.lowercase_required,
            'uppercaseRequired': self.uppercase_required, 'specialRequired':
            self.special_required, 'digitRequired': self.digit_required,
            'minLength': self.min_length, 'maxLength': self.max_length,
            'size': self.size, 'pattern': self.pattern, 'autocomplete':
            self.autocomplete, 'placeholder': self.placeholder, 'disabled':
            self.disabled, 'secret': self.secret, 'errors': self.errors}

    def serialize_value(self) ->str:
        return self.value or ''

    def parse_value(self, value: str) ->str:
        return value



from abstra_internals.controllers.workflow_interface import IWorkflowEngine
from abstra_internals.entities.execution import Execution
from abstra_internals.repositories.factory import Repositories


class DetachedWorkflowEngine(IWorkflowEngine):
    def __init__(self, respositories: Repositories):
        self.repositories = respositories

    def handle_execution_end(self, execution: Execution):
        self.repositories.stage_run.remove(execution.stage_run_id)



from typing import List

from abstra_internals.linter.linter import LinterRule
from abstra_internals.linter.rules.conflicting_path import ConflictingPath
from abstra_internals.linter.rules.duplicate_package_in_requirements import (
    DuplicatePackagesInRequirements,
)
from abstra_internals.linter.rules.env_in_bundle import EnvInBundle
from abstra_internals.linter.rules.missing_abstra_in_requirements import (
    MissingAbstraInRequirements,
)
from abstra_internals.linter.rules.missing_entrypoint import MissingEntrypoint
from abstra_internals.linter.rules.missing_env import MissingEnv
from abstra_internals.linter.rules.missing_packages_in_requirements import (
    MissingPackagesInRequirements,
)
from abstra_internals.linter.rules.new_version_of_abstra_available import (
    NewVersionOfAbstraAvailable,
)
from abstra_internals.linter.rules.syntax_errors import SyntaxErrors
from abstra_internals.linter.rules.unset_get_data import UnsetGetData
from abstra_internals.linter.rules.venv_in_bundle import VenvInBundle

rules: List[LinterRule] = [
    EnvInBundle(),
    VenvInBundle(),
    MissingPackagesInRequirements(),
    SyntaxErrors(),
    MissingEntrypoint(),
    MissingAbstraInRequirements(),
    DuplicatePackagesInRequirements(),
    MissingEnv(),
    NewVersionOfAbstraAvailable(),
    UnsetGetData(),
    ConflictingPath(),
]



from typing import Dict, Optional, Union

from abstra_internals.controllers.execution_store import ExecutionStore


def get_data(key: Optional[str] = None) -> Union[Dict[str, object], object]:
    sdk_ctx = ExecutionStore.get_by_thread()
    return sdk_ctx.workflow_sdk.get_data(key)


def set_data(key: str, value: object):
    sdk_ctx = ExecutionStore.get_by_thread()
    sdk_ctx.workflow_sdk.set_data(key, value)


def set_title(title: str):
    set_data("_thread_title", title)



from typing import Optional

from abstra_internals.credentials import set_credentials
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.usage import editor_manual_usage


def start(credential_token: Optional[str] = None):
    if ProjectRepository.exists():
        print("This directory already has an Abstra project.")
        print("\nRun 'abstra editor' in the directory to start the editor.")
        editor_manual_usage(
            event="run_abstra_start_on_existing_project",
            payload=dict(),
        )
        return

    ProjectRepository.generate_getting_started_project()

    if credential_token:
        set_credentials(credential_token)

    editor_manual_usage(
        event="run_abstra_start_successfully",
        payload=dict(),
    )

    print("Project initialized successfully.")
    print("\nRun 'abstra editor' in the directory to start the editor.")



import abc

from abstra_internals.repositories.project.project import Languages


class BaseTranslation(abc.ABC):
    @abc.abstractmethod
    def confirm_login_at(self, project_name: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def enter_verification_code(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def this_code_expires(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def form_is_waiting(self, stage_name: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def waiting_cta(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def message_from(self, project_name: str) -> str:
        raise NotImplementedError()


class EnglishTranslation(BaseTranslation):
    def confirm_login_at(self, project_name: str) -> str:
        return f"Confirm login at {project_name}"

    def enter_verification_code(self) -> str:
        return "Enter the verification code below in your login screen."

    def this_code_expires(self) -> str:
        return "This verification code expires in 15 minutes."

    def form_is_waiting(self, stage_name: str) -> str:
        return f"{stage_name} is waiting for your input"

    def waiting_cta(self) -> str:
        return "Fill this form"

    def message_from(self, project_name: str) -> str:
        return f"Message from {project_name}"


class PortugueseTranslation(BaseTranslation):
    def confirm_login_at(self, project_name: str) -> str:
        return f"Confirmar login em {project_name}"

    def enter_verification_code(self) -> str:
        return "Digite o código de verificação abaixo na sua tela de login."

    def this_code_expires(self) -> str:
        return "Este código de verificação expira em 15 minutos."

    def form_is_waiting(self, stage_name: str) -> str:
        return f"{stage_name} está esperando seu preenchimento"

    def waiting_cta(self) -> str:
        return "Preencha este formulário"

    def message_from(self, project_name: str) -> str:
        return f"Mensagem de {project_name}"


class GermanTranslation(BaseTranslation):
    def confirm_login_at(self, project_name: str) -> str:
        return f"Anmeldung bestätigen bei {project_name}"

    def enter_verification_code(self) -> str:
        return "Geben Sie den Bestätigungscode unten auf Ihrem Anmeldebildschirm ein."

    def this_code_expires(self) -> str:
        return "Dieser Bestätigungscode läuft in 15 Minuten ab."

    def form_is_waiting(self, stage_name: str) -> str:
        return f"Ein Formular wartet auf Ihre Eingabe in {stage_name}"

    def waiting_cta(self) -> str:
        return "Füllen Sie dieses Formular aus"

    def message_from(self, project_name: str) -> str:
        return f"Nachricht von {project_name}"


class SpanishTranslation(BaseTranslation):
    def confirm_login_at(self, project_name: str) -> str:
        return f"Confirmar inicio de sesión en {project_name}"

    def enter_verification_code(self) -> str:
        return "Ingrese el código de verificación a continuación en su pantalla de inicio de sesión."

    def this_code_expires(self) -> str:
        return "Este código de verificación caduca en 15 minutos."

    def form_is_waiting(self, stage_name: str) -> str:
        return f"{stage_name} está esperando su entrada"

    def waiting_cta(self) -> str:
        return "Complete este formulario"

    def message_from(self, project_name: str) -> str:
        return f"Mensaje de {project_name}"


class FrenchTranslation(BaseTranslation):
    def confirm_login_at(self, project_name: str) -> str:
        return f"Confirmez la connexion à {project_name}"

    def enter_verification_code(self) -> str:
        return "Entrez le code de vérification ci-dessous sur votre écran de connexion."

    def this_code_expires(self) -> str:
        return "Ce code de vérification expire dans 15 minutes."

    def form_is_waiting(self, stage_name: str) -> str:
        return f"{stage_name} attend votre entrée"

    def waiting_cta(self) -> str:
        return "Remplissez ce formulaire"

    def message_from(self, project_name: str) -> str:
        return f"Message de {project_name}"


class HindiTranslation(BaseTranslation):
    def confirm_login_at(self, project_name: str) -> str:
        return f"{project_name} में लॉगिन की पुष्टि करें"

    def enter_verification_code(self) -> str:
        return "अपने लॉगिन स्क्रीन पर नीचे सत्यापन कोड दर्ज करें।"

    def this_code_expires(self) -> str:
        return "यह सत्यापन कोड 15 मिनट में समाप्त हो जाता है।"

    def form_is_waiting(self, stage_name: str) -> str:
        return f"{stage_name} आपकी इनपुट का इंतजार कर रहा है"

    def waiting_cta(self) -> str:
        return "इस फॉर्म को भरें"

    def message_from(self, project_name: str) -> str:
        return f"{project_name} से संदेश"


translations = {
    "en": EnglishTranslation(),
    "pt": PortugueseTranslation(),
    "de": GermanTranslation(),
    "es": SpanishTranslation(),
    "fr": FrenchTranslation(),
    "hi": HindiTranslation(),
}


def get_translation(lang: Languages) -> BaseTranslation:
    return translations[lang]



from typing import Dict, List, Optional

import flask_sock

from abstra_internals.contract import forms_contract
from abstra_internals.contract.forms import BrowserMessageTypes
from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.entities.execution import RequestContext
from abstra_internals.utils import deserialize, serialize


class ClientAbandoned(Exception):
    pass


class FormClient(ExecutionClient):
    def __init__(
        self,
        ws: flask_sock.Server,
        request_context: RequestContext,
        production_mode: bool,
    ) -> None:
        self._request_context = request_context
        self._production_mode = production_mode
        self._ws = ws

    # WEBSOCKET

    def _receive_message(self):
        try:
            str_data = self._ws.receive()
            deserialized = deserialize(str_data)
            if not isinstance(deserialized, dict):
                raise ValueError("Invalid message format")
            return deserialized
        except flask_sock.ConnectionClosed:
            raise ClientAbandoned()

    def _send(self, msg: forms_contract.Message) -> None:
        str_data = serialize(msg.to_json())
        try:
            self._ws.send(str_data)
        except flask_sock.ConnectionClosed:
            pass

    def _user_code_send(self, msg: forms_contract.Message) -> None:
        str_data = serialize(msg.to_json())
        try:
            self._ws.send(str_data)
        except flask_sock.ConnectionClosed:
            raise ClientAbandoned()

    def _wait_for_message(self, *target_types: BrowserMessageTypes) -> Dict:
        ignored_types = ["execution:heartbeat", "debug:close-attempt"]

        while True:
            message = self._receive_message()
            message_type = message.get("type")

            if message_type in ignored_types or message_type not in target_types:
                continue

            return message

    # Inside Target

    ## SDK

    def request_mount_page(
        self,
        widgets: list,
        actions: list,
        end_program: bool,
        reactive_polling_interval: int,
        steps_info: Optional[List],
    ) -> None:
        self._user_code_send(
            forms_contract.FormMountPageMessage(
                widgets,
                actions,
                end_program,
                reactive_polling_interval,
                steps_info,
                self._production_mode,
            )
        )

    def request_page_update(
        self, widgets: list, validation: forms_contract.ValidationResult, event_seq: int
    ) -> None:
        self._user_code_send(
            forms_contract.FormUpdatePageMessage(
                widgets, validation, event_seq, self._production_mode
            )
        )

    def get_query_params(self) -> Dict[str, str]:
        return self._request_context.get("query_params", {})

    def request_auth(self, refresh: bool = False):
        self._user_code_send(
            forms_contract.AuthRequireInfoMessage(refresh, self._production_mode)
        )
        return self._wait_for_message("auth:saved-jwt")

    def handle_invalid_jwt(self):
        self._user_code_send(
            forms_contract.AuthInvalidJWTMessage(self._production_mode)
        )

    def handle_valid_jwt(self):
        self._user_code_send(forms_contract.AuthValidJWTMessage(self._production_mode))

    def request_execute_js(self, code: str, context: dict):
        self._user_code_send(
            forms_contract.ExecuteJSRequestMessage(code, context, self._production_mode)
        )
        data = self._wait_for_message("execute-js:response")
        return data.get("value")

    def request_redirect(
        self, url: str, query_params: Optional[Dict[str, str]]
    ) -> None:
        _query_params = query_params if query_params is not None else {}
        self._user_code_send(forms_contract.RedirectRequestMessage(url, _query_params))

    def next_message(self) -> Dict:
        return self._wait_for_message("form:page-response", "form:user-event")

    ## Target contrioller

    def handle_start(self, execution_id: str):
        self._wait_for_message("execution:start")
        self._send(forms_contract.ExecutionStartedMessage(execution_id))

    def handle_failure(self, e: Exception):
        close_dto = forms_contract.CloseDTO(exit_status="EXCEPTION", exception=e)
        self._send(
            forms_contract.ExecutionEndedMessage(close_dto, self._production_mode)
        )

    def handle_success(self):
        close_dto = forms_contract.CloseDTO(exit_status="SUCCESS")
        self._send(
            forms_contract.ExecutionEndedMessage(close_dto, self._production_mode)
        )

    # Outside target

    ## Execution controller - anomaly

    def handle_unset_thread(self):
        self.handle_failure(Exception("Thread was unset"))

    def handle_lock_failed(self, status: str) -> None:
        self._send(forms_contract.ExecutionLockFailedMessage())



from typing import Dict, TypedDict

from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.entities.execution import RequestContext


class Response(TypedDict):
    headers: Dict[str, str]
    status: int
    body: str


class HookClient(ExecutionClient):
    def __init__(self, request_context: RequestContext) -> None:
        self.response = Response(status=200, body="", headers={})
        self.request_context = request_context

    def handle_lock_failed(self, status: str) -> None:
        self.response["status"] = 423
        pass

    def handle_failure(self, e: Exception) -> None:
        self.response["status"] = 500
        self.response["body"] = "An exception occurred during execution."
        pass

    def handle_success(self) -> None:
        pass

    def set_response(self, status: int, body: str, headers: Dict[str, str]) -> None:
        self.response["headers"] = headers
        self.response["status"] = status
        self.response["body"] = body

    def get_request(self) -> RequestContext:
        return self.request_context

    def handle_start(self, execution_id: str):
        pass

    def handle_unset_thread(self):
        self.response["status"] = 423



import io
from typing import Union
from abstra_internals.widgets.file_utils import upload_widget_file
from abstra_internals.widgets.widget_base import Output


class ImageOutput(Output):
    type = 'image-output'

    def __init__(self, image: Union[str, io.IOBase], **kwargs):
        self.set_props(dict(image=image, **kwargs))

    def set_props(self, props):
        self.image = upload_widget_file(props.get('image', None))
        self.subtitle = props.get('subtitle', '')
        self.full_width = props.get('full_width', False)
        self.label = props.get('label', '')

    def render(self, ctx: dict):
        return {'type': self.type, 'imageUrl': self.image, 'subtitle': self
            .subtitle, 'fullWidth': self.full_width, 'label': self.label}



from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.main import MainController
from abstra_internals.repositories.stage_run import (
    GetStageRunByQueryFilter,
    Pagination,
)


class StageRunsController:
    def __init__(
        self,
        main_controller: MainController,
    ) -> None:
        self.main_controller = main_controller

    def get_leaf_stage_runs(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ):
        return [
            stage_run.to_dto()
            for stage_run in self.main_controller.repositories.stage_run.find_leaves(
                filter=filter, pagination=pagination
            ).stage_runs
        ]

    def get_past_stage_runs(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ):
        return [
            stage_run.to_dto()
            for stage_run in self.main_controller.repositories.stage_run.find_past(
                filter=filter, pagination=pagination
            ).stage_runs
        ]

    def retry(self, stage_run_id: str):
        stage_run = self.main_controller.repositories.stage_run.retry(stage_run_id)
        stage = self.main_controller.get_action(stage_run.stage)
        if stage and stage_run.stage in self.main_controller.get_async_stage_ids():
            ExecutionController(
                repositories=self.main_controller.repositories,
                workflow_engine=self.main_controller.workflow_engine,
            ).submit(stage=stage, target_stage_run_id=stage_run.id)

        return stage_run.to_dto()



import unittest

from abstra_internals.utils.filter import (
    Condition,
    FilterCondition,
    LogicalGroupMultipleConditions,
    LogicalGroupSingleCondition,
    evaluate_condition,
    evaluate_logical_group,
)


class TestEvaluateCondition(unittest.TestCase):
    def test_is_comparator(self):
        self.assertTrue(
            evaluate_condition(Condition("key", "is", "value"), {"key": "value"})
        )
        self.assertFalse(
            evaluate_condition(Condition("key", "is", "value"), {"key": "not_value"})
        )

    def test_contains_comparator_with_string(self):
        self.assertTrue(
            evaluate_condition(
                Condition("key", "contains", "value"), {"key": "This is a value"}
            )
        )
        self.assertFalse(
            evaluate_condition(
                Condition("key", "contains", "value"),
                {"key": "This does not contain the word."},
            )
        )

    def test_contains_comparator_with_not_string(self):
        self.assertFalse(
            evaluate_condition(Condition("key", "contains", "value"), {"key": 123})
        )

    def test_condition_with_missing_key(self):
        self.assertFalse(
            evaluate_condition(Condition("key", "is", "value"), {"other_key": "value"})
        )


class TestEvaluateLogicalGroup(unittest.TestCase):
    def test_and_group(self):
        group = LogicalGroupMultipleConditions(
            operator="AND",
            conditions=[
                Condition("key1", "is", "value1"),
                Condition("key2", "is", "value2"),
            ],
        )
        self.assertTrue(
            evaluate_logical_group(group, {"key1": "value1", "key2": "value2"})
        )
        self.assertFalse(
            evaluate_logical_group(group, {"key1": "value1", "key2": "not_value_2"})
        )

    def test_or_group(self):
        group = LogicalGroupMultipleConditions(
            operator="OR",
            conditions=[
                Condition("key1", "is", "value1"),
                Condition("key2", "is", "value2"),
            ],
        )
        self.assertTrue(
            evaluate_logical_group(group, {"key1": "value1", "key2": "value2"})
        )
        self.assertTrue(
            evaluate_logical_group(group, {"key1": "value1", "key2": "not_value_2"})
        )
        self.assertFalse(
            evaluate_logical_group(
                group, {"key1": "not_value_1", "key2": "not_value_2"}
            )
        )

    def test_not_group(self):
        group = LogicalGroupSingleCondition(
            operator="NOT", condition=Condition("key", "is", "value")
        )
        self.assertFalse(evaluate_logical_group(group, {"key": "value"}))
        self.assertTrue(evaluate_logical_group(group, {"key": "not_value"}))

    def test_group_with_no_conditions(self):
        empty_AND_group = LogicalGroupMultipleConditions(operator="AND", conditions=[])
        empty_OR_group = LogicalGroupMultipleConditions(operator="OR", conditions=[])
        empty_NOT_group = LogicalGroupSingleCondition(operator="NOT", condition=None)

        self.assertTrue(evaluate_logical_group(empty_AND_group, {}))
        self.assertTrue(evaluate_logical_group(empty_OR_group, {}))
        self.assertTrue(evaluate_logical_group(empty_NOT_group, {}))


class TestEvaluateFilterCondition(unittest.TestCase):
    def test_nested_condition(self):
        group = LogicalGroupMultipleConditions(
            operator="AND",
            conditions=[
                LogicalGroupMultipleConditions(
                    operator="OR",
                    conditions=[
                        Condition("key1", "is", "value1"),
                        LogicalGroupSingleCondition(
                            operator="NOT", condition=Condition("key2", "is", "value2")
                        ),
                    ],
                ),
                LogicalGroupMultipleConditions(
                    operator="OR",
                    conditions=[
                        Condition("key2", "is", "value2"),
                        Condition("key3", "is", "value3"),
                    ],
                ),
            ],
        )
        self.assertTrue(
            evaluate_logical_group(
                group, {"key1": "value1", "key2": "not_value_2", "key3": "value3"}
            )
        )
        self.assertTrue(
            evaluate_logical_group(
                group, {"key1": "value1", "key2": "value2", "key3": "value3"}
            )
        )
        self.assertFalse(
            evaluate_logical_group(
                group, {"key1": "not_value_1", "key2": "value2", "key3": "value3"}
            )
        )

    def test_none_is_evaluated_as_true_except_for_not(self):
        group = LogicalGroupMultipleConditions(operator="AND", conditions=[None])
        self.assertTrue(evaluate_logical_group(group, {}))
        group = LogicalGroupMultipleConditions(operator="OR", conditions=[None])
        self.assertTrue(evaluate_logical_group(group, {}))
        group = LogicalGroupSingleCondition(operator="NOT", condition=None)
        self.assertTrue(evaluate_logical_group(group, {}))


class TestFilterCondition(unittest.TestCase):
    def test_from_dict_condition(self):
        data = {"key": "key", "comparator": "is", "value": "value"}
        condition = FilterCondition.from_dict(data)
        self.assertIsInstance(condition, Condition)
        self.assertEqual(condition, Condition("key", "is", "value"))

    def test_from_dict_complex_logic_group(self):
        data = {
            "operator": "AND",
            "conditions": [
                {
                    "operator": "NOT",
                    "condition": {"key": "key1", "comparator": "is", "value": "value1"},
                },
                {
                    "operator": "OR",
                    "conditions": [
                        {"key": "key2", "comparator": "is", "value": "value2"}
                    ],
                },
            ],
        }
        group = FilterCondition.from_dict(data)
        self.assertIsInstance(group, LogicalGroupMultipleConditions)
        self.assertEqual(
            group,
            LogicalGroupMultipleConditions(
                operator="AND",
                conditions=[
                    LogicalGroupSingleCondition(
                        operator="NOT", condition=Condition("key1", "is", "value1")
                    ),
                    LogicalGroupMultipleConditions(
                        operator="OR", conditions=[Condition("key2", "is", "value2")]
                    ),
                ],
            ),
        )

    def test_from_dict_invalid_data(self):
        data = {"invalid": "data"}
        self.assertIsNone(FilterCondition.from_dict(data))

        another_data = {"key": "key", "comparator": "is"}
        self.assertIsNone(FilterCondition.from_dict(another_data))

    def test_to_dict_condition(self):
        condition = Condition("key", "is", "value")
        self.assertEqual(
            condition.to_dict(), {"key": "key", "comparator": "is", "value": "value"}
        )

    def test_to_dict_complex_logic_group(self):
        group = LogicalGroupMultipleConditions(
            operator="AND",
            conditions=[
                LogicalGroupSingleCondition(
                    operator="NOT", condition=Condition("key1", "is", "value1")
                ),
                LogicalGroupMultipleConditions(
                    operator="OR", conditions=[Condition("key2", "is", "value2")]
                ),
            ],
        )
        self.assertEqual(
            group.to_dict(),
            {
                "operator": "AND",
                "conditions": [
                    {
                        "operator": "NOT",
                        "condition": {
                            "key": "key1",
                            "comparator": "is",
                            "value": "value1",
                        },
                    },
                    {
                        "operator": "OR",
                        "conditions": [
                            {"key": "key2", "comparator": "is", "value": "value2"}
                        ],
                    },
                ],
            },
        )



from io import BufferedReader, TextIOWrapper
from pathlib import Path
from typing import List, Optional, Union
from abstra_internals.utils.file import extension_to_mime, upload_file
from abstra_internals.widgets.response_types import FileResponse
from abstra_internals.widgets.widget_base import Input, MultipleHandler


class FileInput(Input):
    type = 'file-input'
    multiple: bool = False
    accepted_formats: list = []
    multiple_handler: MultipleHandler

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)
        self.max_file_size = props.get('max_file_size', None)
        self.multiple = props.get('multiple', False)
        self.min = props.get('min', None)
        self.max = props.get('max', None)
        self.accepted_formats = props.get('accepted_formats', [])
        self.value = props.get('initial_value', self.empty_value)
        self.multiple_handler = MultipleHandler(self.multiple, self.min,
            self.max, required=self.required)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'hint': self.hint,
            'label': self.label, 'value': self.serialize_value(),
            'required': self.required, 'multiple': self.multiple, 'min':
            self.min, 'max': self.max, 'acceptedFormats': self.
            accepted_formats, 'acceptedMimeTypes': self.get_mime_types,
            'fullWidth': self.full_width, 'disabled': self.disabled,
            'maxFileSize': self.max_file_size, 'errors': self.errors}

    @staticmethod
    def __get_file_uri(value: Union[FileResponse, str, BufferedReader,
        TextIOWrapper]) ->str:
        if isinstance(value, str):
            if Path(value).is_file():
                return upload_file(open(value))
            else:
                return value
        if isinstance(value, FileResponse):
            return value._url
        if isinstance(value, BufferedReader):
            return upload_file(value)
        if isinstance(value, TextIOWrapper):
            return upload_file(value)
        return ''

    def serialize_value(self) ->List[str]:
        values_list = self.multiple_handler.value_to_list(self.value)
        return [FileInput.__get_file_uri(item) for item in values_list]

    def parse_value(self, value: Optional[List[str]]) ->Union[FileResponse,
        List[FileResponse], None]:
        file_responses = [FileResponse(url) for url in value or []]
        return self.multiple_handler.value_to_list_or_value(file_responses)

    def validate(self) ->List[str]:
        errors = []
        errors.extend(self.multiple_handler.validate(self.value))
        if not self.value:
            return errors
        if isinstance(self.value, list):
            for file in self.value:
                errors.extend(self._validate_file_extension(file))
        else:
            errors.extend(self._validate_file_extension(self.value))
        return errors

    @property
    def accepted_formats_without_dot(self) ->List[str]:
        return [extension.replace('.', '') for extension in self.
            accepted_formats]

    def _validate_file_extension(self, file: FileResponse) ->List[str]:
        if not self.accepted_formats:
            return []
        file_extension = file._url.split('.')[-1]
        if file_extension not in self.accepted_formats_without_dot:
            return ['i18n_error_invalid_file_format']
        return []

    @property
    def get_mime_types(self) ->str:
        if not self.accepted_formats:
            return '*'
        return ','.join(extension_to_mime(ext) for ext in self.accepted_formats
            )



import abc
import typing

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import SIDECAR_HEADERS


class TablesApiHttpClient(abc.ABC):
    execute_url: str

    def __init__(self, base_url: str) -> None:
        self.execute_url = f"{base_url}/tables/execute"

    def execute(self, query: str, params: typing.List) -> requests.Response:
        raise NotImplementedError()


class ProductionTablesApiHttpClient(TablesApiHttpClient):
    def execute(self, query: str, params: typing.List) -> requests.Response:
        body = {"query": query, "params": params}
        return requests.post(self.execute_url, headers=SIDECAR_HEADERS, json=body)


class LocalTablesApiHttpClient(TablesApiHttpClient):
    def execute(self, query: str, params: typing.List) -> requests.Response:
        body = {"query": query, "params": params}
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to execute a table query")
        return requests.post(self.execute_url, headers=headers, json=body)



import threading
import time
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv
from watchdog.events import FileModifiedEvent, FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from abstra_internals.modules import reload_project_local_modules
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.settings import Settings

IGNORED_FILES = [".abstra/resources.dat", ".abstra/", "abstra.json"]
DEBOUNCE_DELAY = 0.5


class FileChangeEventHandler(FileSystemEventHandler):
    def __init__(self):
        super().__init__()
        self._debounce_timer: Optional[threading.Timer] = None

    def on_modified(self, event: FileSystemEvent):
        has_ignored_paths = any(
            ignored_file in str(event.src_path) for ignored_file in IGNORED_FILES
        )

        if not isinstance(event, FileModifiedEvent) or has_ignored_paths:
            return

        if event.src_path.endswith(".env"):
            load_dotenv(Settings.root_path / ".env", override=True)
            return self.reload_files()

        for dep in ProjectRepository.load().get_local_dependencies():
            if dep.resolve() == Path(event.src_path).resolve():
                return self.reload_files()

    def reload_files(self):
        if self._debounce_timer is not None:
            self._debounce_timer.cancel()

        self._debounce_timer = threading.Timer(
            DEBOUNCE_DELAY, reload_project_local_modules
        )
        self._debounce_timer.start()


def watch_file_change_events():
    event_handler = FileChangeEventHandler()

    observer = Observer()
    observer.schedule(event_handler, path=str(Settings.root_path), recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



from pathlib import Path

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.controllers.workflow_engine import WorkflowEngine
from abstra_internals.controllers.workflow_engine_detached import DetachedWorkflowEngine
from abstra_internals.entities.execution import RequestContext
from abstra_internals.repositories.project.project import HookStage, ProjectRepository
from tests.fixtures import BaseTest


class ExecutionControllerTest(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.workflow_engine = WorkflowEngine(self.repositories)

        self.detached_workflow_engine = DetachedWorkflowEngine(self.repositories)

        self.request_context = RequestContext(
            body={"a": 1}.__str__(),
            headers={"auth": "some_secret_token"},
            query_params={"c": "3"},
            method="GET",
        )

        self.project = ProjectRepository.load()
        self.stage = HookStage.create(
            title="mock_stage",
            file="mock_file.py",
            id="mock_hook_id",
            workflow_position=(0, 1),
        )
        Path(self.stage.file).write_text("print('Hello, World!')")
        self.project.add_stage(self.stage)
        ProjectRepository.save(self.project)

        self.hook_client = HookClient(self.request_context)

    def test_run_initial_returns_dto(self):
        ExecutionController(
            repositories=self.repositories, workflow_engine=self.workflow_engine
        ).run(
            stage=self.stage,
            client=self.hook_client,
            request=self.request_context,
        )

        self.assertEqual(self.hook_client.response["status"], 200)

    def test_run_detached_returns_dto(self):
        ExecutionController(
            repositories=self.repositories,
            workflow_engine=self.detached_workflow_engine,
        ).test(
            stage=self.stage,
            client=self.hook_client,
            request=self.request_context,
        )

        self.assertEqual(self.hook_client.response["status"], 200)



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

import requests

from abstra_internals.contracts_generated import CommonUser, CommonUserRoles
from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import SIDECAR_HEADERS


class UsersRepository(ABC):
    @abstractmethod
    def get_user(self, email: str) -> Optional[CommonUser]:
        pass

    @abstractmethod
    def insert_user(self, email: str) -> bool:
        pass


class TestUsersRepository(UsersRepository):
    ## prevent pytest to inspect this class
    __test__ = False
    data: Dict[str, CommonUserRoles]

    def __init__(self, data: Dict[str, CommonUserRoles] = {}) -> None:
        super().__init__()
        self.data = data

    def get_user(self, email: str) -> Optional[CommonUser]:
        if email in self.data:
            return CommonUser(
                id=email,
                email=email,
                roles=self.data[email],
                created_at="2021-09-01T00:00:00Z",
                project_id="1",
            )

        return None

    def insert_user(self, email: str) -> bool:
        if email in self.data:
            return False
        self.data[email] = []
        return True


class LocalUsersRepository(UsersRepository):
    @property
    def headers(self):
        return resolve_headers()

    def get_user(self, email: str) -> Optional[CommonUser]:
        return None

    def insert_user(self, email: str) -> bool:
        return True


class ProductionUsersRepository(UsersRepository):
    def __init__(self, url: str) -> None:
        self.url = url

    def _request(
        self,
        method: str,
        path: str,
        body: Any = None,
        params: dict = {},
        raise_for_status: bool = True,
    ):
        r = requests.request(
            method=method,
            url=f"{self.url}/users{path}",
            headers=SIDECAR_HEADERS,
            json=body,
            params=params,
        )

        if raise_for_status:
            r.raise_for_status()

        return r

    def get_user(self, email: str) -> Optional[CommonUser]:
        r = self._request("GET", "/", params={"email": email}, raise_for_status=False)
        if not r.ok:
            return None
        return CommonUser.from_dict(r.json())

    def insert_user(self, email: str) -> bool:
        r = self._request("POST", "/", body={"email": email}, raise_for_status=False)
        return r.ok



import flask

from abstra_internals.contracts_generated import (
    AbstraLibApiEditorLintersFixResponse,
)
from abstra_internals.controllers.linters import check_linters, fix_linter
from abstra_internals.usage import editor_usage


def get_editor_bp():
    bp = flask.Blueprint("editor_linters", __name__)

    # 1s pooling in this route
    @bp.get("/check")
    def _check_linters():
        return check_linters()

    @bp.post("/fix/<rule_name>/<fix_name>")
    @editor_usage
    def _fix_linter(rule_name: str, fix_name: str):
        fix_linter(rule_name, fix_name)
        return AbstraLibApiEditorLintersFixResponse(success=True).to_dict()

    return bp



from typing import List, Union

from abstra_internals.controllers.execution_store import ExecutionStore
from abstra_internals.email_templates import message_template


def send_email(to: Union[str, List[str]], message: str, title: str = ""):
    if isinstance(to, str):
        to = [to]

    ExecutionStore.get_by_thread().repositories.email.send(
        message_template.generate_email(to, message, title)
    )


__all__ = [
    "send_email",
]



import signal
import typing as t

from abstra_internals.logger import AbstraLogger


class SignalHandlers:
    initialized = False

    @classmethod
    def init(cls):
        if cls.initialized:
            return

        signal.signal(signal.SIGTERM, cls.handle_sigterm)

        cls.initialized = True

    sigterm_callbacks: t.List[t.Callable] = []

    @classmethod
    def register_sigterm_callback(cls, callback: t.Callable):
        cls.sigterm_callbacks.append(callback)

    @classmethod
    def handle_sigterm(cls, signum, frame):
        AbstraLogger.warning("[SIGNALS] SIGTERM received")
        for callback in cls.sigterm_callbacks:
            callback()

    @classmethod
    def clear(cls):
        cls.sigterm_callbacks.clear()



from abstra_internals.widgets.widget_base import Input


class TextInput(Input):
    type = 'text-input'
    empty_value = ''

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', '')
        self.required = props.get('required', True)
        self.value = props.get('initial_value', self.empty_value)
        self.placeholder = props.get('placeholder', '')
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.mask = props.get('mask', None)
        self.max_length = props.get('max_length', None)
        self.min_length = props.get('min_length', None)
        self.disabled = props.get('disabled', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'value': self.serialize_value(), 'placeholder': self.
            placeholder, 'required': self.required, 'hint': self.hint,
            'fullWidth': self.full_width, 'mask': self.mask, 'maxLength':
            self.max_length, 'minLength': self.min_length, 'disabled': self
            .disabled, 'errors': self.errors}

    def serialize_value(self) ->str:
        return self.value or ''

    def validate(self):
        validation_errors = super().validate()
        if self.max_length and len(self.value) > self.max_length:
            validation_errors.append(
                f'Value is too long. Max length is {self.max_length}')
        if self.min_length and len(self.value) < self.min_length:
            validation_errors.append(
                f'Value is too short. Min length is {self.min_length}')
        return validation_errors



import abstra.hooks as ah
import abstra.workflows as aw

# Use Abstra Hooks to create Python endpoints
body, query, headers = ah.get_request()

print("⚙️ Hook is running... received body:", body)

# You can store data in the current thread,
# so it will be available in the next stages
aw.set_data("hook_data", body)

# You can send a response back to the client after doing some processing
ah.send_json({"message": "A message from the hook!"})



from typing import Optional

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution import RequestContext
from abstra_internals.environment import set_SERVER_UUID, set_WORKER_UUID
from abstra_internals.logger import AbstraLogger, Environment
from abstra_internals.repositories.project.project import ActionStage
from abstra_internals.settings import Settings
from abstra_internals.stdio_patcher import StdioPatcher


# runs in subprocess - all arguments must be picklable/multiprocessable
def ExecutionProcess(
    *,
    root_path: str,
    worker_uuid: str,
    arbiter_uuid: str,
    stage: ActionStage,
    controller: MainController,
    environment: Optional[Environment],
    request: Optional[RequestContext] = None,
    target_stage_run_id: Optional[str] = None,
):
    Settings.set_root_path(root_path)
    AbstraLogger.init(environment)

    set_WORKER_UUID(worker_uuid)
    set_SERVER_UUID(arbiter_uuid)
    StdioPatcher.apply(controller)

    head_id = worker_uuid.split("-")[0]

    AbstraLogger.debug(f"[{head_id}] WORKER INIT")

    try:
        ExecutionController(
            repositories=controller.repositories,
            workflow_engine=controller.workflow_engine,
        ).run(
            stage=stage,
            request=request,
            target_stage_run_id=target_stage_run_id,
        )
    except Exception as e:
        AbstraLogger.error(f"[{head_id}] WORKER ERROR: {e}")
        AbstraLogger.capture_exception(e)
    finally:
        AbstraLogger.debug(f"[{head_id}] WORKER EXIT")



from typing import List

import flask

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.kanban import DataRequest, KanbanController
from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.workflows import get_path, make_stage_dto
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.repositories.execution_logs import LogEntry
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.repositories.stage_run import StageRun
from abstra_internals.server.guards.role_guard import Guard, StageIdSelector
from abstra_internals.usage import player_usage


def get_editor_bp(main_controller: MainController):
    controller = KanbanController(
        main_controller.stage_run_repository, main_controller.execution_logs_repository
    )
    bp = flask.Blueprint("kanban", __name__)

    @bp.get("/stages")
    def _get_stages():
        stages = controller._get_stages()
        return [stage.to_dict() for stage in stages]

    # 1s pooling in this route
    @bp.post("/")
    def _get_kanban():
        if flask.request.json is None:
            flask.abort(400)
        req = DataRequest.from_dict(flask.request.json)
        return controller.get_data(req).to_dict()

    @bp.post("/count")
    def _count_by_status():
        return controller.count_by_status()

    @bp.get("/path")
    def _get_path():
        n = int(flask.request.args.get("n", 3))
        return get_path(n)

    # 1s pooling in this route
    @bp.get("/logs/<stage_run_id>")
    def _get_ancestor_logs(stage_run_id: str):
        project = ProjectRepository.load()

        def kanban_log_filter(log: LogEntry):
            return (isinstance(log, LogEntry)) and log.payload["text"].strip() != ""

        def entry_from_run(stage_run: StageRun, logs: List[LogEntry]):
            stage = project.get_stage(stage_run.stage)

            return {
                "stage_run": stage_run.to_dto(),
                "stage": make_stage_dto(stage) if stage else None,
                "logs": [log.to_dto() for log in logs if kanban_log_filter(log)],
            }

        return [
            entry_from_run(stage_run, logs)
            for stage_run, logs in controller.get_ancestor_logs(stage_run_id)
        ]

    @bp.post("/jobs/<path:id>/start")
    def _start_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        ExecutionController(
            repositories=main_controller.repositories,
            workflow_engine=main_controller.workflow_engine,
        ).submit(stage=job)

        return {"status": "running"}

    return bp


def get_player_bp(main_controller: MainController):
    guard = Guard(main_controller.users_repository, enabled=IS_PRODUCTION)

    controller = KanbanController(
        main_controller.stage_run_repository,
        main_controller.execution_logs_repository,
        read_only=True,
    )
    bp = flask.Blueprint("kanban", __name__)

    # 1s pooling in this route
    @bp.get("/stages")
    @guard.by(StageIdSelector("kanban"))
    def _get_stages():
        stages = controller._get_stages()
        return [stage.to_dict() for stage in stages]

    @bp.get("/path")
    @guard.by(StageIdSelector("kanban"))
    def _get_path():
        n = int(flask.request.args.get("n", 3))
        return get_path(n)

    # 1s pooling in this route
    @bp.post("/")
    @guard.by(StageIdSelector("kanban"))
    def _get_kanban():
        if flask.request.json is None:
            flask.abort(400)
        req = DataRequest.from_dict(flask.request.json)
        return controller.get_data(req).to_dict()

    # 1s pooling in this route
    @bp.post("/count")
    @guard.by(StageIdSelector("kanban"))
    def _count_by_status():
        return controller.count_by_status()

    @bp.get("/logs/<stage_run_id>")
    @guard.by(StageIdSelector("kanban"))
    @player_usage
    def _get_ancestor_logs(stage_run_id: str):
        project = ProjectRepository.load()

        def entry_from_run(stage_run: StageRun, logs: List[LogEntry]):
            stage = project.get_stage(stage_run.stage)

            return {
                "stage_run": stage_run.to_dto(),
                "stage": make_stage_dto(stage) if stage else None,
                "logs": [log.to_dto() for log in logs],
            }

        return [
            entry_from_run(stage_run, logs)
            for stage_run, logs in controller.get_ancestor_logs(stage_run_id)
        ]

    @bp.post("/jobs/<path:id>/start")
    @guard.by(StageIdSelector("kanban"))
    @player_usage
    def _start_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        ExecutionController(
            repositories=main_controller.repositories,
            workflow_engine=main_controller.workflow_engine,
        ).submit(stage=job)

        return {"status": "running"}

    return bp
